import click
from dotenv import load_dotenv

from likeasrt.commands import run
from likeasrt.domain.azurespeech import generate_srt_from_file
from likeasrt.logs import get_app_logger

load_dotenv()

logger = get_app_logger()


@click.command(name="gen")
@click.option("--source", "-s", help="source audio file", required=True)
@click.option(
    "--language",
    "-l",
    help="source audio file language",
    default="en-US",
    required=True,
)
@click.option(
    "--write-transcript",
    "-t",
    help="if specified, a transcript file is generated from the extracted information",
    default=False,
    required=False,
    is_flag=True,
)
@click.option(
    "--write-json",
    "-j",
    default=False,
    help=(
        "if specified, it writes JSON output of each "
        "Azure Speech recognition event on disk."
    ),
    is_flag=True,
)
def generate_srt_command(
    source: str, language: str, write_transcript: bool, write_json: bool
):
    """
    Given an input audio file, it obtains a transcript using Azure Speech, then
    generates a subtitles file having a name equal to the one of the input audio file,
    but ".srt" extension.
    """
    run(lambda: generate_srt_from_file(source, language, write_transcript, write_json))



import json
import os
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union

import azure.cognitiveservices.speech as speechsdk
from essentials.folders import ensure_folder, get_file_extension

from likeasrt.domain.core import SpokenWord, get_segments
from likeasrt.domain.srt import get_related_file_name, write_segments
from likeasrt.errors import (
    InputFileNotFoundError,
    InvalidInputError,
    MissingEnvVariableError,
)
from likeasrt.logs import get_app_logger

logger = get_app_logger()


class RecognitionError(Exception):
    def __init__(self) -> None:
        super().__init__("A portion was not recognized. Cannot continue the process.")


@dataclass
class SpeechSettings:
    speech_subscription: str
    speed_endpoint: str

    @classmethod
    def from_env(cls):
        try:
            return cls(os.environ["SPEECH_SUBSCRIPTION"], os.environ["SPEECH_ENDPOINT"])
        except KeyError as key_error:
            raise MissingEnvVariableError(key_error.args[0]) from key_error


def tick_to_seconds(value: int) -> float:
    # A single tick represents one hundred nanoseconds or one ten-millionth of a second.
    # See: https://bit.ly/3talWxK
    return value / 1e7


def parse_result(raw_details: str) -> dict:
    return json.loads(raw_details)


def _eng_transform(display_text: str) -> str:
    return display_text.replace("i'm", "I'm")


def get_words(raw_details: str) -> List[SpokenWord]:
    details = parse_result(raw_details)
    n_best = details["NBest"]
    best_result = n_best[0]

    words = []

    display = _eng_transform(best_result.get("Display", ""))
    display_words_with_punctuation: List[str] = display.split(" ")

    for index, word in enumerate(best_result.get("Words", [])):
        # Offset and Duration are expressed in ticks, each tick is 100 nanoseconds
        start_time_in_seconds = tick_to_seconds(word["Offset"])
        end_time_in_seconds = start_time_in_seconds + tick_to_seconds(word["Duration"])

        text = word["Word"]
        # try to restore punctuation
        if display_words_with_punctuation:
            try:
                corresponding_display_word = display_words_with_punctuation[index]

                if text.lower() in corresponding_display_word.lower():
                    text = corresponding_display_word
            except IndexError:
                pass

        words.append(
            SpokenWord(
                start_time=start_time_in_seconds,
                end_time=end_time_in_seconds,
                text=text,
            )
        )

    return words


def extract_recognition(
    file_path: Union[str, Path], language: str = "en-US"
) -> List[speechsdk.SpeechRecognitionEventArgs]:
    settings = SpeechSettings.from_env()

    speech_config = speechsdk.SpeechConfig(
        subscription=settings.speech_subscription,
        endpoint=settings.speed_endpoint,
    )

    # necessary to generate .srt
    speech_config.request_word_level_timestamps()

    audio_input = speechsdk.AudioConfig(filename=str(file_path))
    speech_recognizer = speechsdk.SpeechRecognizer(
        speech_config=speech_config, audio_config=audio_input, language=language
    )

    speech_recognizer.start_continuous_recognition()

    done = False

    events = []

    def on_recognized(evt):
        if evt.result.reason != speechsdk.ResultReason.RecognizedSpeech:
            raise RecognitionError()

        logger.info("RECOGNIZED: %s", evt)
        events.append(evt)

    def stop_cb(evt):
        logger.info("Recognition process complete...")
        speech_recognizer.stop_continuous_recognition()
        nonlocal done
        done = True

    speech_recognizer.recognized.connect(on_recognized)
    speech_recognizer.session_stopped.connect(stop_cb)

    while not done:
        time.sleep(0.5)

    return events


def generate_from_events(
    events: List[speechsdk.SpeechRecognitionEventArgs], output_file_name: str
) -> None:
    all_words = []

    for evt in events:
        all_words += get_words(evt.result.json)

    segments = get_segments(all_words)
    srt = "".join(write_segments(segments))

    logger.info("Writing output file %s", output_file_name)
    with open(output_file_name, mode="wt", encoding="utf8") as output_srt_file:
        output_srt_file.write(srt)


def write_transcript_file(
    events: List[speechsdk.SpeechRecognitionEventArgs],
    output_file_name: str = "transcript.txt",
) -> None:
    with open(output_file_name, mode="wt", encoding="utf8") as output_file:
        for evt in events:
            output_file.write(evt.result.text + "\n")


def write_events_data(
    events: List[speechsdk.SpeechRecognitionEventArgs], file_names_prefix: str = ""
) -> None:
    """
    Writes detailed event data
    """
    out_folder = Path("out")
    ensure_folder("out")

    for index, evt in enumerate(events):
        with open(
            out_folder / f"{file_names_prefix}-evt-{index}.json",
            encoding="utf8",
            mode="wt",
        ) as evt_file:
            evt_file.write(
                json.dumps(json.loads(evt.result.json), ensure_ascii=False, indent=4)
            )


def validate_audio_file_path(file_path: Union[str, Path]) -> None:
    source_file = Path(file_path)

    extension = get_file_extension(source_file)

    if extension.lower() != ".wav":
        raise InvalidInputError(
            "The source file must be of type wav. Refer to the documentation for a "
            "reference on how to convert audio files using ffmpeg."
        )

    if not source_file.exists():
        raise InputFileNotFoundError(file_path)


def generate_srt_from_file(
    file_path: Union[str, Path],
    language: str = "en-US",
    write_transcript: bool = False,
    write_json: bool = False,
) -> None:
    validate_audio_file_path(file_path)
    output_file_name = get_related_file_name(file_path, ".srt")

    logger.info("Obtaining data from Azure Speech...")
    events = extract_recognition(file_path, language)

    if write_transcript:
        logger.info("Writing a transcript file: %s", output_file_name)
        transcript_file = get_related_file_name(file_path, "-transcript.txt")
        write_transcript_file(events, transcript_file)

    if write_json:
        logger.info(
            "Writing detailed JSON output from Azure Speech on disk, in 'out' folder."
        )
        write_events_data(events, get_related_file_name(file_path, ""))

    logger.debug("Writing the .srt output file: %s", output_file_name)
    generate_from_events(events, output_file_name)



"""
Functions and methods that are specific to SubRip .srt format.
https://en.wikipedia.org/wiki/SubRip
"""
from pathlib import Path
from typing import Iterable, Union

from essentials.folders import split_path

from .core import Segment


def write_time(seconds: float) -> str:
    hours, milliseconds = divmod(seconds * 1000, 3600000)
    minutes, milliseconds = divmod(milliseconds, 60000)
    seconds = float(milliseconds) / 1000
    return ("%02i:%02i:%06.3f" % (hours, minutes, seconds)).replace(".", ",")


def write_segment(segment: Segment) -> str:
    """
    Example:
        00:00:00,240 --> 00:00:03,580
        Hello! How are you?
    """
    return (
        f"{write_time(segment.start_time)} --> {write_time(segment.end_time)}\n"
        + segment.text
        + "\n"
    )


def write_segments(segments: Iterable[Segment]) -> Iterable[str]:
    for index, segment in enumerate(segments):
        yield str(index) + "\n"
        yield write_segment(segment) + "\n"


def get_related_file_name(
    original_file_path: Union[str, Path], suffix: str = ".srt"
) -> str:
    """
    Returns a new file name, starting from an original file and applying the given
    suffix. The suffix is supposed to contain the result file extension!
    """
    _, file_name, _ = split_path(str(original_file_path))
    return file_name + suffix



import logging
import sys

import click

from . import VERSION
from .commands.srt import generate_srt_command
from .logs import get_app_logger

sys.path.append(".")


@click.group()
@click.option(
    "--verbose", default=False, help="Whether to display debug output.", is_flag=True
)
@click.version_option(version=VERSION)
def main(verbose):
    """
    CLI to generate subtitles in srt format from audio files, using Azure Speech.

    Roberto Prevato https://github.com/RobertoPrevato/Like-a-srt
    """
    logger = get_app_logger()
    if verbose:
        logger.setLevel(logging.DEBUG)

    logger.debug("Running in --verbose mode")


main.add_command(generate_srt_command)



from dataclasses import dataclass
from typing import Iterable, List, Optional


@dataclass
class SpokenWord:
    start_time: float
    end_time: float
    text: str


class SegmentWithoutWordsError(TypeError):
    def __init__(self) -> None:
        super().__init__("This segment does not contain any word")


@dataclass
class Segment:
    "Represents a segment in subtitles."
    words: List[SpokenWord]

    def add_word(self, word: SpokenWord) -> None:
        self.words.append(word)

    @property
    def start_time(self) -> float:
        if not self.words:
            raise SegmentWithoutWordsError()
        return self.words[0].start_time

    @property
    def end_time(self) -> float:
        if not self.words:
            raise SegmentWithoutWordsError()
        return self.words[-1].end_time

    @property
    def text(self) -> str:
        return " ".join(word.text for word in self.words)


def get_segments(
    spoken_words: Iterable[SpokenWord], segment_length: float = 3.0
) -> Iterable[Segment]:
    current_segment: Optional[Segment] = None

    for word in spoken_words:
        if current_segment is None:
            current_segment = Segment([word])
        else:
            if word.end_time - current_segment.start_time > segment_length:
                # create a new segment
                yield current_segment
                current_segment = Segment([word])
            else:
                # add the word to the current segment
                current_segment.add_word(word)

    if current_segment is not None:
        yield current_segment



from pathlib import Path
from typing import List, Union


class CLIError(Exception):
    """
    Base class for all exception risen by the CLI.
    """


class InvalidInputError(CLIError):
    """
    Generic exception type for invalid input.
    """


class InputFileNotFoundError(InvalidInputError):
    """
    Exception risen when an expected input file is not found.
    """

    def __init__(self, file_path: Union[str, Path]) -> None:
        super().__init__(f"File {file_path} not found")


class MissingEnvVariableError(CLIError):
    """
    Exception risen when an expected environmental variable is not configured.
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Missing ENV variable: {name}")


class MissingEnvVariablesError(CLIError):
    """
    Exception risen when expected environmental variables are not configured.
    """

    def __init__(self, names: List[str]) -> None:
        super().__init__(f"Missing ENV variables: {', '.join(names)}")



import logging
import logging.handlers
from datetime import datetime

from essentials.folders import ensure_folder

logger = None


def get_app_logger():
    global logger

    if logger is not None:
        return logger

    logger = logging.getLogger("app")

    logger.setLevel(logging.INFO)

    max_bytes = 24 * 1024 * 1024

    file_handler = logging.handlers.RotatingFileHandler

    now = datetime.now()
    ts = now.strftime("%Y%m%d")

    ensure_folder(f"logs/{ts}")

    handler = file_handler(f"logs/{ts}/app.log", maxBytes=max_bytes, backupCount=5)

    handler.setLevel(logging.DEBUG)

    logger.addHandler(handler)
    logger.addHandler(logging.StreamHandler())

    return logger



import sys

from likeasrt.errors import CLIError, InvalidInputError
from likeasrt.logs import get_app_logger

logger = get_app_logger()


def run(function):
    """
    Runs a function for a command, handling common exceptions such as KeyboardInterrupt,
    CLI errors, invalid input, etc.
    """
    try:
        function()
    except InvalidInputError as invalid_input:
        logger.error("Error: %s", invalid_input)
        sys.exit(1)
    except CLIError as cli_error:
        logger.error(cli_error)
        sys.exit(2)
    except KeyboardInterrupt:
        logger.info("[*] User interrupted")
        sys.exit(1)



from setuptools import setup

from likeasrt import VERSION


def readme():
    with open("README.md", encoding="utf8") as readme_file:
        return readme_file.read()


setup(
    name="like-a-srt",
    version=VERSION,
    description=(
        "CLI to generate SRT subtitles automatically from audio files, "
        "using Azure Speech"
    ),
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/RobertoPrevato/Like-a-srt",
    author="RobertoPrevato",
    author_email="roberto.prevato@gmail.com",
    keywords="azure speech srt subtitles automatic generation",
    license="MIT",
    packages=[
        "likeasrt",
        "likeasrt.commands",
        "likeasrt.domain",
    ],
    entry_points={
        "console_scripts": ["like-a-srt=likeasrt.main:main", "las=likeasrt.main:main"]
    },
    install_requires=[
        "click==8.0.3",
        "essentials==1.1.4",
        "azure-cognitiveservices-speech==1.19.0",
    ],
    include_package_data=True,
)
